import asyncio
import logging

from ..plugin_base import JanusPlugin
from aiortc import RTCPeerConnection, RTCSessionDescription, VideoStreamTrack
from .media import MediaPlayer

logger = logging.getLogger(__name__)


class JanusVideoRoomPlugin(JanusPlugin):
    """Janus VideoRoom plugin instance

    Implements API to interact with VideoRoom plugin.

    Each plugin instance is expected to have only 1 PeerConnection
    """

    name = "janus.plugin.videoroom"  #: Plugin name
    pc: RTCPeerConnection

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.joined_event = asyncio.Event()
        self.pc = RTCPeerConnection()
        # self.loop = asyncio.get_running_loop()

    async def on_receive(self, response: dict):
        if response["janus"] == "event":
            logger.info(f"Event response: {response}")
            if "plugindata" in response:
                if response["plugindata"]["data"]["videoroom"] == "attached":
                    # Subscriber attached
                    self.joined_event.set()
                elif response["plugindata"]["data"]["videoroom"] == "joined":
                    # Participant joined (joined as publisher but may not publish)
                    self.joined_event.set()
        else:
            logger.info(f"Unimplemented response handle: {response['janus']}")
            logger.info(response)

        # Handle JSEP. Could be answer or offer.
        if "jsep" in response:
            asyncio.create_task(self.handle_jsep(response["jsep"]))

    async def join(self, room_id: int, publisher_id: int, display_name: str) -> None:
        """Join a room

        | A handle can join a room and then do nothing, but this should be called before publishing.
        | There is an API to configure and publish at the same time, but it's not implemented yet.

        :param room_id: Room ID to join. This must be available at the server.
        :param publisher_id: Your publisher ID to set.
        :param display_name: Your display name when you join the room.
        """

        message_transaction = await self.send(
            {
                "janus": "message",
                "body": {
                    "request": "join",
                    "ptype": "publisher",
                    "room": room_id,
                    "id": publisher_id,
                    "display": display_name,
                },
            },
        )
        response = await message_transaction.get(
            {
                "janus": "event",
                "plugindata": {
                    "plugin": "janus.plugin.videoroom",
                    "data": {"videoroom": "joined"},
                },
            }
        )
        await message_transaction.done()
        logger.info(f"Room join response: {response}")

    async def leave(self) -> None:
        message_transaction = await self.send(
            {
                "janus": "message",
                "body": {
                    "request": "leave",
                },
            }
        )
        response = await message_transaction.get(
            {
                "janus": "event",
                "plugindata": {
                    "plugin": "janus.plugin.videoroom",
                    "data": {"videoroom": "event", "leaving": "ok"},
                },
            }
        )
        await message_transaction.done()
        logger.info(f"Room leave response: {response}")

    async def publish(self, ffmpeg_input, width: int, height: int) -> None:
        """Publish video stream to the room

        Should already have joined a room before this. Then this will publish the
        video stream to the handle.
        """

        # create media source
        player = MediaPlayer(
            ffmpeg_input,
            width,
            height,
        )
        # Just save the media player. Not used
        self.player = player

        # configure media
        media = {"audio": False, "video": True}
        if player and player.audio:
            self.pc.addTrack(player.audio)
            media["audio"] = True

        if player and player.video:
            self.pc.addTrack(player.video)
        else:
            self.pc.addTrack(VideoStreamTrack())

        # send offer
        await self.pc.setLocalDescription(await self.pc.createOffer())

        request = {"request": "configure"}
        request.update(media)

        message_transaction = await self.send(
            {
                "janus": "message",
                "body": request,
                "jsep": {
                    "sdp": self.pc.localDescription.sdp,
                    "trickle": False,
                    "type": self.pc.localDescription.type,
                },
            }
        )
        await message_transaction.get()
        await message_transaction.done()

        await self.joined_event.wait()

    async def unpublish(self) -> None:
        """Stop publishing"""

        message_transaction = await self.send(
            {
                "janus": "message",
                "body": {
                    "request": "unpublish",
                },
            }
        )
        await message_transaction.get()
        await message_transaction.done()
        await self.pc.close()

    async def subscribe(self, room_id: int, feed_id: int) -> None:
        """Subscribe to a feed

        :param room_id: Room ID containing the feed. The same ID that
            you would use to join the room.
        :param feed_id: ID of the feed that you want to stream. Should be their publisher ID.
        """

        message_transaction = await self.send(
            {
                "janus": "message",
                "body": {
                    "request": "join",
                    "ptype": "subscriber",
                    "room": room_id,
                    "feed": feed_id,
                    # "close_pc": True,
                    # "audio": True,
                    # "video": True,
                    # "data": True,
                    # "offer_audio": True,
                    # "offer_video": True,
                    # "offer_data": True,
                },
            }
        )
        await message_transaction.get()
        await message_transaction.done()
        await self.joined_event.wait()

    async def unsubscribe(self) -> None:
        """Unsubscribe from the feed"""

        message_transaction = await self.send(
            {
                "janus": "message",
                "body": {
                    "request": "leave",
                },
            }
        )
        await message_transaction.get()
        await message_transaction.done()
        self.joined_event.clear()

    async def start(self, answer=None) -> None:
        """Signal WebRTC start. I guess"""

        payload = {"janus": "message", "body": {"request": "start"}}
        if answer:
            payload["jsep"] = {
                "sdp": answer,
                "type": "answer",
                "trickle": True,
            }
        message_transaction = await self.send(payload)
        await message_transaction.get()
        await message_transaction.done()

    async def pause(self) -> None:
        """Pause media streaming"""

        message_transaction = await self.send(
            {
                "janus": "message",
                "body": {
                    "request": "pause",
                },
            }
        )
        await message_transaction.get()
        await message_transaction.done()

    async def list_participants(self, room_id: int) -> list:
        """Get participant list

        Get a list of publishers in the room, that are currently publishing.

        :param room_id: List participants in this room
        :return: A list containing the participants. Can be empty.
        """

        message_transaction = await self.send(
            {
                "janus": "message",
                "body": {
                    "request": "listparticipants",
                    "room": room_id,
                },
            }
        )
        response = await message_transaction.get()
        await message_transaction.done()
        return response["plugindata"]["data"]["participants"]

    async def handle_jsep(self, jsep):
        logger.info(jsep)
        if "sdp" in jsep:
            sdp = jsep["sdp"]
            if jsep["type"] == "answer":
                logger.info(f"Received answer:\n{sdp}")

                # apply answer
                await self.pc.setRemoteDescription(
                    RTCSessionDescription(sdp=jsep["sdp"], type=jsep["type"])
                )
            elif jsep["type"] == "offer":
                pass
            else:
                raise Exception("Invalid JSEP")

        elif "ice" in jsep:
            ice = jsep["ice"]
            candidate = ice["candidate"]
            sdpMLineIndex = ice["sdpMLineIndex"]
            self.webrtcbin.emit("add-ice-candidate", sdpMLineIndex, candidate)



import logging
import asyncio
from dataclasses import dataclass
from typing import Dict

import aiohttp

from .transport import JanusTransport


logger = logging.getLogger(__name__)


@dataclass
class ReceiverTask:
    task: asyncio.Task
    destroyed_event: asyncio.Event


class JanusTransportHTTP(JanusTransport):
    """Janus transport through HTTP"""

    __receive_response_task_map: Dict[int, ReceiverTask]
    __api_secret: str
    __token: str

    def __init__(
        self, base_url: str, api_secret: str = None, token: str = None, **kwargs: dict
    ):
        super().__init__(base_url=base_url, api_secret=api_secret, token=token)

        self.__receive_response_task_map = dict()
        # HTTP transport needs these for long polling
        self.__api_secret = api_secret
        self.__token = token

    async def _connect(self):
        pass

    async def _disconnect(self):
        pass

    def __build_url(self, session_id: int = None, handle_id: int = None) -> str:
        url = f"{self.base_url}"

        if session_id:
            url = f"{url}/{session_id}"

            if handle_id:
                url = f"{url}/{handle_id}"

        return url

    async def info(self) -> Dict:
        async with aiohttp.ClientSession() as http_session:
            async with http_session.get(f"{self.base_url}/info") as response:
                return await response.json()

    async def _send(
        self,
        message: Dict,
    ) -> None:
        session_id = message.get("session_id")
        handle_id = message.get("handle_id")

        async with aiohttp.ClientSession() as http_session:
            async with http_session.post(
                url=self.__build_url(session_id=session_id, handle_id=handle_id),
                json=message,
            ) as response:
                response.raise_for_status()

                response_dict = await response.json()

                # if "error" in response_dict:
                #     raise Exception(response_dict)

                # # There must be a transaction ID
                # response_transaction_id = response_dict["transaction"]

                # Fake receive
                # # We will immediately get a response in the HTTP response, so need
                # # to put this into the queue
                # await self.put_response(
                #     transaction_id=response_transaction_id, response=response_dict
                # )
                await self.receive(response=response_dict)

    def session_receive_response_done_cb(
        self, task: asyncio.Task, context=None
    ) -> None:
        try:
            # Check if any exceptions are raised
            task.exception()
        except asyncio.CancelledError:
            logger.info("Receive message task ended")
        except asyncio.InvalidStateError:
            logger.info("receive_message_done_cb called with invalid state")
        except Exception as err:
            logger.error(err)

    async def session_receive_response(
        self, session_id: str, destroyed_event: asyncio.Event
    ) -> None:
        url_params = {}
        if self.__api_secret:
            url_params["apisecret"] = self.__api_secret
        if self.__token:
            url_params["token"] = self.__token

        async with aiohttp.ClientSession() as http_session:
            while not destroyed_event.is_set():
                async with http_session.get(
                    url=self.__build_url(session_id=session_id),
                    params=url_params,
                ) as response:
                    # Maybe session is destroyed during http request
                    if destroyed_event.is_set():
                        break

                    response.raise_for_status()

                    response_dict = await response.json()

                    if "error" in response_dict:
                        raise Exception(response_dict)

                    if response_dict["janus"] == "keepalive":
                        continue

                    await self.receive(response=response_dict)

    async def dispatch_session_created(self, session_id: str) -> None:
        logger.info(f"Create session_receive_response task ({session_id})")
        destroyed_event = asyncio.Event()
        task = asyncio.create_task(
            self.session_receive_response(
                session_id=session_id, destroyed_event=destroyed_event
            )
        )
        task.add_done_callback(self.session_receive_response_done_cb)
        self.__receive_response_task_map[session_id] = ReceiverTask(
            task=task, destroyed_event=destroyed_event
        )

    async def dispatch_session_destroyed(self, session_id: int) -> None:
        if session_id not in self.__receive_response_task_map:
            logger.warn(f"Session receive response task not found for {session_id}")

        logger.info(f"Destroy session_receive_response task ({session_id})")
        receiver_task = self.__receive_response_task_map[session_id]
        # Don't use task.cancel() to avoid
        # Exception ignored in: <function _ProactorBasePipeTransport.__del__ at 0x0000027A269465F0>
        receiver_task.destroyed_event.set()

        # Destroying sessions could cost some time because it needs to
        # wait for the long-poll request to complete
        await asyncio.wait([receiver_task.task])


def protocol_matcher(base_url: str):
    return base_url.startswith(("http://", "https://"))


JanusTransport.register_transport(
    protocol_matcher=protocol_matcher, transport_cls=JanusTransportHTTP
)



import logging
import asyncio

from aiortc import (
    RTCPeerConnection,
    RTCSessionDescription,
    # VideoStreamTrack,
    MediaStreamTrack,
)
from aiortc.contrib.media import MediaPlayer, MediaRecorder

from .plugin_base import JanusPlugin
from .message_transaction import is_subset

logger = logging.getLogger(__name__)


class JanusVideoCallPlugin(JanusPlugin):
    """Janus Video Call plugin implementation"""

    name = "janus.plugin.videocall"  #: Plugin name
    __username: str
    __pc: RTCPeerConnection
    __player: MediaPlayer
    __recorder: MediaRecorder

    def __init__(self) -> None:
        super().__init__()

        self.__username = ""
        self.__pc = None
        self.__player = None
        self.__recorder = None

    async def on_receive(self, response: dict):
        # Handle JSEP. Could be answer or offer.
        if "jsep" in response and response["jsep"]["type"] == "answer":
            await self.on_receive_jsep(response["jsep"])

        janus_code = response["janus"]

        if janus_code == "media":
            if response["receiving"]:
                # It's ok to start multiple times, only the track that
                # has not been started will start
                await self.__recorder.start()

        if janus_code == "event":
            logger.info(f"Event response: {response}")
            if "plugindata" in response:
                if response["plugindata"]["data"]["videocall"] == "event":
                    event_result = response["plugindata"]["data"]["result"]
                    logger.info(f"Event result: {event_result}")
                    if (
                        "event" in event_result
                        and event_result["event"] == "incomingcall"
                    ):
                        asyncio.create_task(
                            self.on_incoming_call(plugin=self, jsep=response["jsep"])
                        )
        else:
            logger.info(f"Unimplemented response handle: {response}")

    async def on_receive_jsep(self, jsep: dict):
        if self.__pc and self.__pc.signalingState != "closed":
            await self.__pc.setRemoteDescription(
                RTCSessionDescription(sdp=jsep["sdp"], type=jsep["type"])
            )

    async def create_pc(
        self, player: MediaPlayer, recorder: MediaRecorder = None, jsep: dict = {}
    ) -> RTCPeerConnection:
        pc = RTCPeerConnection()

        # configure media
        if player.audio:
            pc.addTrack(player.audio)

        if player.video:
            pc.addTrack(player.video)

        # Must configure on track event before setRemoteDescription
        if recorder:

            @pc.on("track")
            async def on_track(track: MediaStreamTrack):
                logger.info("Track %s received" % track.kind)
                if track.kind == "video":
                    recorder.addTrack(track)
                if track.kind == "audio":
                    recorder.addTrack(track)

        if jsep:
            await pc.setRemoteDescription(
                RTCSessionDescription(sdp=jsep["sdp"], type=jsep["type"])
            )

        return pc

    async def on_incoming_call(self, jsep: dict):
        """Override this. This will be called when plugin receive incoming call event"""
        pass
        # # self.__player = MediaPlayer("./Into.the.Wild.2007.mp4")
        # self.__player = MediaPlayer("http://download.tsi.telecom-paristech.fr/gpac/dataset/dash/uhd/mux_sources/hevcds_720p30_2M.mp4")
        # self.__recorder = MediaRecorder("./videocall_in_record.mp4")
        # self.__pc = await self.create_pc(
        #     player=self.__player,
        #     recorder=self.__recorder,
        #     jsep=jsep,
        # )

        # await self.__pc.setLocalDescription(await self.__pc.createAnswer())
        # jsep = {
        #     "sdp": self.__pc.localDescription.sdp,
        #     "trickle": False,
        #     "type": self.__pc.localDescription.type,
        # }
        # await self.accept(jsep=jsep)

    async def send_wrapper(self, message: dict, matcher: dict, jsep: dict = {}) -> dict:
        def function_matcher(message: dict):
            return is_subset(message, matcher) or is_subset(
                message,
                {
                    "janus": "event",
                    "plugindata": {
                        "plugin": "janus.plugin.videocall",
                        "data": {
                            "videocall": "event",
                            "error_code": None,
                            "error": None,
                        },
                    },
                },
            )

        full_message = message
        if jsep:
            full_message = {**message, "jsep": jsep}

        message_transaction = await self.send(
            message=full_message,
        )
        response = await message_transaction.get(matcher=function_matcher)
        await message_transaction.done()

        return response

    async def list(self) -> list:
        response = await self.send_wrapper(
            message={
                "janus": "message",
                "body": {
                    "request": "list",
                },
            },
            matcher={
                "janus": "event",
                "plugindata": {
                    "plugin": "janus.plugin.videocall",
                    "data": {"videocall": "event", "result": {"list": None}},
                },
            },
        )

        return response["plugindata"]["data"]["result"]["list"]

    async def register(self, username: str) -> bool:
        """Register a username

        Detach plugin to de-register the username
        """
        if self.__username:
            raise Exception(f"Can only register 1 username: {self.__username}")

        matcher_success = {
            "janus": "event",
            "plugindata": {
                "plugin": "janus.plugin.videocall",
                "data": {
                    "videocall": "event",
                    "result": {
                        "event": "registered",
                    },
                },
            },
        }
        response = await self.send_wrapper(
            message={
                "janus": "message",
                "body": {
                    "request": "register",
                    "username": username,
                },
            },
            matcher=matcher_success,
        )

        if is_subset(response, matcher_success):
            self.__username = username
            return True
        else:
            return False

    async def call(
        self, username: str, player: MediaPlayer, recorder: MediaRecorder = None
    ) -> bool:
        if not self.__username:
            raise Exception("Register a username first")
        self.__pc = await self.create_pc(player=player, recorder=recorder)
        self.__player = player
        self.__recorder = recorder

        # send offer
        await self.__pc.setLocalDescription(await self.__pc.createOffer())

        jsep = {
            "sdp": self.__pc.localDescription.sdp,
            "trickle": True,
            "type": self.__pc.localDescription.type,
        }

        matcher_success = {
            "janus": "event",
            "plugindata": {
                "plugin": "janus.plugin.videocall",
                "data": {
                    "videocall": "event",
                    "result": {
                        "event": "calling",
                    },
                },
            },
        }
        response = await self.send_wrapper(
            message={
                "janus": "message",
                "body": {
                    "request": "call",
                    "username": username,
                },
            },
            matcher=matcher_success,
            jsep=jsep,
        )

        return is_subset(response, matcher_success)

    async def accept(
        self,
        jsep: dict,
        pc: RTCPeerConnection,
        player: MediaPlayer,
        recorder: MediaRecorder = None,
    ) -> bool:
        self.__pc = pc
        self.__player = player
        self.__recorder = recorder

        matcher_success = {
            "janus": "event",
            "plugindata": {
                "plugin": "janus.plugin.videocall",
                "data": {
                    "videocall": "event",
                    "result": {
                        "event": "accepted",
                    },
                },
            },
        }
        response = await self.send_wrapper(
            message={
                "janus": "message",
                "body": {
                    "request": "accept",
                },
            },
            matcher=matcher_success,
            jsep=jsep,
        )

        return is_subset(response, matcher_success)

    async def set(self, audio: bool, video: bool, jsep: dict = {}) -> bool:
        matcher_success = {
            "janus": "event",
            "plugindata": {
                "plugin": "janus.plugin.videocall",
                "data": {
                    "videocall": "event",
                    "result": {
                        "event": "set",
                    },
                },
            },
        }
        body = {
            "request": "set",
            "audio": audio,
            "video": video,
            # "bitrate" : <numeric bitrate value>,
            # "record" : true|false,
            # "filename" : <base path/filename to use for the recording>,
            # "substream" : <substream to receive (0-2), in case simulcasting is enabled>,
            # "temporal" : <temporal layers to receive (0-2), in case simulcasting is enabled>,
            # "fallback" : <How much time (in us, default 250000) without receiving packets will make us drop to the substream below>
        }
        response = await self.send_wrapper(
            message={
                "janus": "message",
                "body": body,
            },
            matcher=matcher_success,
            jsep=jsep,
        )

        return is_subset(response, matcher_success)

    async def hangup(self) -> bool:
        matcher_success = {
            "janus": "event",
            "plugindata": {
                "plugin": "janus.plugin.videocall",
                "data": {
                    "videocall": "event",
                    "result": {
                        "event": "hangup",
                    },
                },
            },
        }
        response = await self.send_wrapper(
            message={
                "janus": "message",
                "body": {
                    "request": "hangup",
                },
            },
            matcher=matcher_success,
        )

        # Stream ended. Ok to close PC multiple times.
        if self.__pc:
            await self.__pc.close()
            self.__pc = None
        # Ok to stop recording multiple times.
        if self.__recorder:
            await self.__recorder.stop()
            self.__recorder = None
        self.__player = None

        return is_subset(response, matcher_success)



import asyncio
import logging
import threading
from typing import Optional, Set, Union
import subprocess
import time

from av import VideoFrame
from av.frame import Frame
from av.packet import Packet
import numpy as np
import fractions

from aiortc.mediastreams import MediaStreamError, MediaStreamTrack

logger = logging.getLogger(__name__)

REAL_TIME_FORMATS = [
    "alsa",
    "android_camera",
    "avfoundation",
    "bktr",
    "decklink",
    "dshow",
    "fbdev",
    "gdigrab",
    "iec61883",
    "jack",
    "kmsgrab",
    "openal",
    "oss",
    "pulse",
    "sndio",
    "rtsp",
    "v4l2",
    "vfwcap",
    "x11grab",
]


async def blackhole_consume(track):
    while True:
        try:
            await track.recv()
        except MediaStreamError:
            return


class MediaBlackhole:
    """
    A media sink that consumes and discards all media.
    """

    def __init__(self):
        self.__tracks = {}

    def addTrack(self, track):
        """
        Add a track whose media should be discarded.

        :param track: A :class:`aiortc.MediaStreamTrack`.
        """
        if track not in self.__tracks:
            self.__tracks[track] = None

    async def start(self):
        """
        Start discarding media.
        """
        for track, task in self.__tracks.items():
            if task is None:
                self.__tracks[track] = asyncio.ensure_future(blackhole_consume(track))

    async def stop(self):
        """
        Stop discarding media.
        """
        for task in self.__tracks.values():
            if task is not None:
                task.cancel()
        self.__tracks = {}


class PlayerStreamTrack(MediaStreamTrack):
    def __init__(self, player, kind):
        super().__init__()
        self.kind = kind
        self._player = player
        self._queue = asyncio.Queue()
        self._start = None

    async def recv(self) -> Union[Frame, Packet]:
        if self.readyState != "live":
            raise MediaStreamError

        self._player._start(self)
        data = await self._queue.get()
        if data is None:
            self.stop()
            raise MediaStreamError
        # data_time = None

        # # control playback rate
        # if (
        #     self._player is not None
        #     and self._player._throttle_playback
        #     and data_time is not None
        # ):
        #     if self._start is None:
        #         self._start = time.time() - data_time
        #     else:
        #         wait = self._start + data_time - time.time()
        #         await asyncio.sleep(wait)

        return data

    def stop(self):
        super().stop()
        if self._player is not None:
            self._player._stop(self)
            self._player = None


class MediaPlayer:
    """
    A media source that reads audio and/or video from a file.

    Examples:

    .. code-block:: python

        # Open a video file.
        player = MediaPlayer('/path/to/some.mp4')

        # Open an HTTP stream.
        player = MediaPlayer(
            'http://download.tsi.telecom-paristech.fr/'
            'gpac/dataset/dash/uhd/mux_sources/hevcds_720p30_2M.mp4')

        # Open webcam on Linux.
        player = MediaPlayer('/dev/video0', format='v4l2', options={
            'video_size': '640x480'
        })

        # Open webcam on OS X.
        player = MediaPlayer('default:none', format='avfoundation', options={
            'video_size': '640x480'
        })

        # Open webcam on Windows.
        player = MediaPlayer('video=Integrated Camera', format='dshow', options={
            'video_size': '640x480'
        })

    :param file: The path to a file, or a file-like object.
    :param format: The format to use, defaults to autodect.
    :param options: Additional options to pass to FFmpeg.
    :param timeout: Open/read timeout to pass to FFmpeg.
    :param loop: Whether to repeat playback indefinitely (requires a seekable file).
    """

    def __init__(
        self,
        ffmpeg_input,
        width: int,
        height: int,
    ):
        self.__thread: Optional[threading.Thread] = None
        self.__thread_quit: Optional[threading.Event] = None

        self.__ffmpeg_input = ffmpeg_input
        self.__width = width
        self.__height = height

        # examine streams
        self.__started: Set[MediaStreamTrack] = set()
        self.__audio: Optional[PlayerStreamTrack] = None
        self.__video = PlayerStreamTrack(self, kind="video")

        self._throttle_playback = False
        self._loop_playback = False

    @property
    def audio(self) -> MediaStreamTrack:
        """
        A :class:`aiortc.MediaStreamTrack` instance if the file contains audio.
        """
        return self.__audio

    @property
    def video(self) -> MediaStreamTrack:
        """
        A :class:`aiortc.MediaStreamTrack` instance if the file contains video.
        """
        return self.__video

    def player_worker_decode(
        self,
        loop,
        video_track,
        quit_event,
    ):
        logging.info("Thread 1: starting")
        pts = 0
        video_process: subprocess.Popen = self.__ffmpeg_input.output(
            "pipe:", format="rawvideo", pix_fmt="rgb24"
        ).run_async(pipe_stdout=True, pipe_stdin=True)

        count = 0
        total_time = 0
        total_cpu_time = 0
        while not quit_event.is_set():
            in_bytes = video_process.stdout.read(self.__width * self.__height * 3)
            if not in_bytes:
                break

            start_time = time.time()
            start_cpu_time = time.process_time()

            in_frame = np.frombuffer(in_bytes, np.uint8).reshape(
                [self.__height, self.__width, 3]
            )
            frame = VideoFrame.from_ndarray(in_frame, format="rgb24")

            frame.pts = pts
            pts += 1
            frame.time_base = fractions.Fraction(1, 48000)

            total_time += time.time() - start_time
            total_cpu_time += time.process_time() - start_cpu_time
            count += 1

            # logging.info(frame)
            asyncio.run_coroutine_threadsafe(video_track._queue.put(frame), loop)

        video_process.communicate(b"q", timeout=1)
        video_process.wait()

        logging.info("Thread 1: finishing")
        logger.info(
            f"time: {total_time * 1000 / count}ms | cpu time: {total_cpu_time * 1000 / count}ms"
        )

    def _start(self, track: PlayerStreamTrack) -> None:
        self.__started.add(track)
        if self.__thread is None:
            self.__log_debug("Starting worker thread")
            self.__thread_quit = threading.Event()
            self.__thread = threading.Thread(
                name="media-player",
                target=self.player_worker_decode,
                args=(
                    asyncio.get_event_loop(),
                    self.__video,
                    self.__thread_quit,
                ),
            )
            self.__thread.start()

    def _stop(self, track: PlayerStreamTrack) -> None:
        self.__started.discard(track)

        if not self.__started and self.__thread is not None:
            self.__log_debug("Stopping worker thread")
            self.__thread_quit.set()
            self.__thread.join()
            self.__thread = None

        # if not self.__started and self.__container is not None:
        #     self.__container.close()
        #     self.__container = None

    def __log_debug(self, msg: str, *args) -> None:
        logger.debug(f"MediaPlayer(%s) {msg}", "asd", *args)




from .admin_monitor import JanusAdminMonitorClient
from .session import JanusSession, PluginAttachFail

from .plugin_base import JanusPlugin
from .plugin_echotest import JanusEchoTestPlugin
from .plugin_video_call import JanusVideoCallPlugin
from .plugin_video_room import JanusVideoRoomPlugin

from .transport import JanusTransport
from .transport_http import JanusTransportHTTP
from .transport_websocket import JanusTransportWebsocket

import logging
logging.getLogger("janus_client").addHandler(logging.NullHandler())



import asyncio
import uuid
from typing import Dict, List, Union, Callable


def is_subset(dict_1: Dict, dict_2: Dict) -> bool:
    """Check if dict_2 is subset of dict_1 recursively

    Only checks dict, str or int type in dict_2
    """
    if not isinstance(dict_1, dict):
        raise TypeError(f"dict_1 must be a dictionary: {dict_1}")

    if not isinstance(dict_2, dict):
        raise TypeError(f"dict_2 must be a dictionary: {dict_2}")

    if not dict_2:
        return True

    for key_2, val_2 in dict_2.items():
        if not (
            isinstance(val_2, dict) or isinstance(val_2, str) or isinstance(val_2, int)
        ):
            # If not these few types, then only need
            # key_2 to be in dict_1
            if key_2 in dict_1:
                continue
            else:
                return False

        # Need to check values
        val_1 = dict_1.get(key_2, None)

        # Simple compare
        if val_1 == val_2:
            continue

        # Now val_2 can be another dict
        if isinstance(val_1, dict) and isinstance(val_2, dict):
            if is_subset(val_1, val_2):
                continue
            else:
                return False

        # key_2: val_2 is not in dict_1, so False
        return False

    # All of dict_2 is found in dict_1
    return True


class MessageTransaction:
    __id: str
    __msg_all: List[Dict]
    __msg_in: asyncio.Queue

    def __init__(self) -> None:
        self.__id = uuid.uuid4().hex
        self.__msg_all = []
        self.__msg_in = asyncio.Queue()

    @property
    def id(self) -> str:
        return self.__id

    def put_msg(self, message: Dict) -> None:
        # Queue is never full
        self.__msg_in.put_nowait(message)

    async def get(
        self,
        matcher: Union[Dict, Callable] = lambda *args, **kwargs: True,
        timeout: Union[float, None] = None,
    ) -> Dict:
        if not (isinstance(matcher, dict) or callable(matcher)):
            raise TypeError(f"matcher must be callable or dictionary: {matcher}")

        _matcher: Callable
        if callable(matcher):
            # matcher is a function
            _matcher = matcher
        else:
            # matcher is a dict
            def dict_matcher(msg: dict) -> bool:
                return is_subset(msg, matcher)

            _matcher = dict_matcher

        # Try to find message in saved messages
        for msg in self.__msg_all:
            if _matcher(msg):
                return msg

        # Wait in queue until a matching message is found
        msg = await asyncio.wait_for(self.__msg_in.get(), timeout=timeout)
        # Always save received messages
        self.__msg_all.append(msg)

        while not _matcher(msg):
            msg = await asyncio.wait_for(self.__msg_in.get(), timeout=timeout)
            self.__msg_all.append(msg)

        return msg

    async def on_done(self) -> None:
        pass

    async def done(self) -> None:
        """Must call this when finish using to release resources"""
        await self.on_done()




from ..plugin_base import JanusPlugin
import asyncio
import logging
logger = logging.getLogger(__name__)

import gi
gi.require_version('GLib', '2.0')
gi.require_version('GObject', '2.0')
gi.require_version('Gst', '1.0')
from gi.repository import Gst
gi.require_version('GstWebRTC', '1.0')
from gi.repository import GstWebRTC
gi.require_version('GstSdp', '1.0')
from gi.repository import GstSdp

# Set to False to send H.264
DO_VP8 = True
# Set to False to disable RTX (lost packet retransmission)
DO_RTX = True
# Choose the video source:
VIDEO_SRC="videotestsrc pattern=ball"
# VIDEO_SRC = "v4l2src"

if DO_VP8:
    (encoder, payloader, rtp_encoding) = (
        "vp8enc target-bitrate=100000 overshoot=25 undershoot=100 deadline=33000 keyframe-max-dist=1", "rtpvp8pay picture-id-mode=2", "VP8")
else:
    (encoder, payloader, rtp_encoding) = ("x264enc",
                                          "rtph264pay aggregate-mode=zero-latency", "H264")

PIPELINE_DESC = '''
 webrtcbin name=sendrecv stun-server=stun://stun.l.google.com:19302
 {} ! video/x-raw,width=640,height=480 ! videoconvert ! queue !
 {} ! {} !  queue ! application/x-rtp,media=video,encoding-name={},payload=96 ! sendrecv.
'''.format(VIDEO_SRC, encoder, payloader, rtp_encoding)


class JanusVideoRoomPlugin(JanusPlugin):
    """Janus VideoRoom plugin instance

    Implements API to interact with VideoRoom plugin.
    """

    name = "janus.plugin.videoroom" #: Plugin name

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.joined_event = asyncio.Event()
        self.gst_webrtc_ready = asyncio.Event()
        self.loop = asyncio.get_running_loop()

        # Create a new pipeline, elements will be added to this.
        self.pipeline = Gst.Pipeline.new()

    def handle_async_response(self, response: dict):
        if response["janus"] == "event":
            logger.info(f"Event response: {response}")
            if "plugindata" in response:
                if response["plugindata"]["data"]["videoroom"] == "attached":
                    # Subscriber attached
                    self.joined_event.set()
                elif response["plugindata"]["data"]["videoroom"] == "joined":
                    # Participant joined (joined as publisher but may not publish)
                    self.joined_event.set()
        else:
            logger.info(f"Unimplemented response handle: {response['janus']}")
            logger.info(response)
        # Handle JSEP. Could be answer or offer.
        if "jsep" in response:
            asyncio.create_task(self.handle_jsep(response["jsep"]))

    async def join(self, room_id: int, publisher_id: int, display_name: str) -> None:
        """Join a room

        | A handle can join a room and then do nothing, but this should be called before publishing.
        | There is an API to configure and publish at the same time, but it's not implemented yet.

        :param room_id: Room ID to join. This must be available at the server.
        :param publisher_id: Your publisher ID to set.
        :param display_name: Your display name when you join the room.
        """

        await self.send({
            "janus": "message",
            "body": {
                "request": "join",
                "ptype": "publisher",
                "room": room_id,
                "id": publisher_id,
                "display": display_name,
            },
        })
        await self.joined_event.wait()

    async def publish(self) -> None:
        """Publish some hardcoded video stream to the handle

        Should already have joined a room before this. Then this will publish the
        hardcoded video stream to the handle.
        """

        # Initialize Gst WebRTC
        self.start_pipeline()
        await self.gst_webrtc_ready.wait()
        # Create offer
        promise = Gst.Promise.new()
        self.webrtcbin.emit('create-offer', None, promise)
        promise.wait()
        reply = promise.get_reply()
        offer = reply.get_value('offer')
        # Set local description
        promise = Gst.Promise.new()
        self.webrtcbin.emit('set-local-description', offer, promise)
        promise.interrupt()

        text = offer.sdp.as_text()
        logger.info('Sending offer and publishing:\n%s' % text)
        await self.send({
            "janus": "message",
            "body": {
                "request": "publish",
                "audio": True,
                "video": True,
            },
            "jsep": {
                'sdp': text,
                'type': 'offer',
                'trickle': True,
            }
        })
        await self.joined_event.wait()

    async def unpublish(self) -> None:
        """Stop publishing"""

        logger.info("Set pipeline to null")
        self.pipeline.set_state(Gst.State.NULL)
        logger.info("Set pipeline complete")
        await self.send({
            "janus": "message",
            "body": {
                "request": "unpublish",
            }
        })
        self.gst_webrtc_ready.clear()

    async def subscribe(self, room_id: int, feed_id: int) -> None:
        """Subscribe to a feed

        :param room_id: Room ID containing the feed. The same ID that
            you would use to join the room.
        :param feed_id: ID of the feed that you want to stream. Should be their publisher ID.
        """

        # Create webrtcbin element named app
        self.webrtcbin = Gst.ElementFactory.make("webrtcbin", "app")
        self.webrtcbin.connect('on-negotiation-needed',
                            self.on_negotiation_needed)
        self.webrtcbin.connect('on-ice-candidate',
                            self.send_ice_candidate_message)
        self.webrtcbin.connect('pad-added', self.on_incoming_stream)
        self.pipeline.add(self.webrtcbin)
        # trans = self.webrtcbin.emit('get-transceiver', 0)
        # if DO_RTX:
        #     trans.set_property('do-nack', True)
        self.pipeline.set_state(Gst.State.PLAYING)
        await self.send({
            "janus": "message",
            "body": {
                "request": "join",
                "ptype": "subscriber",
                "room": room_id,
                "feed": feed_id,
                # "close_pc": True,
                # "audio": True,
                # "video": True,
                # "data": True,
                # "offer_audio": True,
                # "offer_video": True,
                # "offer_data": True,
            }
        })
        await self.joined_event.wait()

    async def unsubscribe(self) -> None:
        """Unsubscribe from the feed"""

        self.pipeline.set_state(Gst.State.NULL)
        await self.send({
            "janus": "message",
            "body": {
                "request": "leave",
            }
        })
        self.joined_event.clear()
        self.gst_webrtc_ready.clear()

    async def start(self, answer=None) -> None:
        """Signal WebRTC start. I guess"""

        payload = {
            "janus": "message",
            "body": {
                "request": "start"
            }
        }
        if answer:
            payload["jsep"] = {
                'sdp': answer,
                'type': 'answer',
                'trickle': True,
            }
        await self.send(payload)

    async def pause(self) -> None:
        """Pause media streaming"""

        await self.send({
            "janus": "message",
            "body": {
                "request": "pause",
            }
        })

    async def list_participants(self, room_id: int) -> list:
        """Get participant list

        Get a list of publishers in the room, that are currently publishing.

        :param room_id: List participants in this room
        :return: A list containing the participants. Can be empty.
        """

        response = await self.send({
            "janus": "message",
            "body": {
                "request": "listparticipants",
                "room": room_id,
            }
        })
        return response["plugindata"]["data"]["participants"]

    def on_negotiation_needed(self, element):
        logger.info("on_negotiation_needed called")
        self.gst_webrtc_ready.set()
        # promise = Gst.Promise.new_with_change_func(self.on_offer_created, element, None)
        # element.emit('create-offer', None, promise)

    def send_ice_candidate_message(self, _, sdpMLineIndex, candidate):
        icemsg = {'candidate': candidate, 'sdpMLineIndex': sdpMLineIndex}
        logger.info (f"Sending ICE {icemsg}")
        # loop = asyncio.new_event_loop()
        future = asyncio.run_coroutine_threadsafe(
            self.trickle(sdpMLineIndex, candidate), self.loop)
        future.result()

    def on_incoming_decodebin_stream(self, _, pad):
        if not pad.has_current_caps():
            logger.info(pad, 'has no caps, ignoring')
            return

        caps = pad.get_current_caps()
        name = caps.to_string()
        if name.startswith('video'):
            q = Gst.ElementFactory.make('queue')
            conv = Gst.ElementFactory.make('videoconvert')
            sink = Gst.ElementFactory.make('autovideosink')
            self.pipeline.add(q)
            self.pipeline.add(conv)
            self.pipeline.add(sink)
            self.pipeline.sync_children_states()
            pad.link(q.get_static_pad('sink'))
            q.link(conv)
            conv.link(sink)
        elif name.startswith('audio'):
            q = Gst.ElementFactory.make('queue')
            conv = Gst.ElementFactory.make('audioconvert')
            resample = Gst.ElementFactory.make('audioresample')
            sink = Gst.ElementFactory.make('autoaudiosink')
            self.pipeline.add(q)
            self.pipeline.add(conv)
            self.pipeline.add(resample)
            self.pipeline.add(sink)
            self.pipeline.sync_children_states()
            pad.link(q.get_static_pad('sink'))
            q.link(conv)
            conv.link(resample)
            resample.link(sink)

    def on_incoming_stream(self, _, pad):
        if pad.direction != Gst.PadDirection.SRC:
            return

        decodebin = Gst.ElementFactory.make('decodebin')
        decodebin.connect('pad-added', self.on_incoming_decodebin_stream)
        self.pipeline.add(decodebin)
        decodebin.sync_state_with_parent()
        self.webrtcbin.link(decodebin)

    def start_pipeline(self):
        self.pipeline = Gst.parse_launch(PIPELINE_DESC)
        self.webrtcbin = self.pipeline.get_by_name('sendrecv')
        self.webrtcbin.connect('on-negotiation-needed',
                            self.on_negotiation_needed)
        self.webrtcbin.connect('on-ice-candidate',
                            self.send_ice_candidate_message)
        self.webrtcbin.connect('pad-added', self.on_incoming_stream)

        trans = self.webrtcbin.emit('get-transceiver', 0)
        if DO_RTX:
            trans.set_property('do-nack', True)
        self.pipeline.set_state(Gst.State.PLAYING)

    def extract_ice_from_sdp(self, sdp):
        mlineindex = -1
        for line in sdp.splitlines():
            if line.startswith("a=candidate"):
                candidate = line[2:]
                if mlineindex < 0:
                    logger.info("Received ice candidate in SDP before any m= line")
                    continue
                logger.info(
                    'Received remote ice-candidate mlineindex {}: {}'.format(mlineindex, candidate))
                self.webrtcbin.emit('add-ice-candidate', mlineindex, candidate)
            elif line.startswith("m="):
                mlineindex += 1

    async def handle_jsep(self, jsep):
        logger.info(jsep)
        if 'sdp' in jsep:
            sdp = jsep['sdp']
            if jsep['type'] == 'answer':
                logger.info('Received answer:\n%s' % sdp)
                _, sdpmsg = GstSdp.SDPMessage.new()
                GstSdp.sdp_message_parse_buffer(bytes(sdp.encode()), sdpmsg)

                answer = GstWebRTC.WebRTCSessionDescription.new(
                    GstWebRTC.WebRTCSDPType.ANSWER, sdpmsg)
                promise = Gst.Promise.new()
                self.webrtcbin.emit('set-remote-description', answer, promise)
                promise.interrupt()

                # Extract ICE candidates from the SDP to work around a GStreamer
                # limitation in (at least) 1.16.2 and below
                # This is tested to be not needed on 1.19.0.1
                # self.extract_ice_from_sdp(sdp)
            elif jsep['type'] == 'offer':
                logger.info('Received offer:\n%s' % sdp)
                _, sdpmsg = GstSdp.SDPMessage.new()
                GstSdp.sdp_message_parse_buffer(bytes(sdp.encode()), sdpmsg)

                offer = GstWebRTC.WebRTCSessionDescription.new(
                    GstWebRTC.WebRTCSDPType.OFFER, sdpmsg)
                promise = Gst.Promise.new()
                self.webrtcbin.emit('set-remote-description', offer, promise)
                promise.interrupt()

                # Extract ICE candidates from the SDP to work around a GStreamer
                # limitation in (at least) 1.16.2 and below
                self.extract_ice_from_sdp(sdp)

                # direction = GstWebRTC.WebRTCRTPTransceiverDirection.RECVONLY
                # caps = Gst.caps_from_string("application/x-rtp,media=video,encoding-name=VP8/9000,payload=96")
                # self.webrtcbin.emit('add-transceiver', direction, caps)

                # Create answer
                promise = Gst.Promise.new()
                self.webrtcbin.emit('create-answer', None, promise)
                promise.wait()
                reply = promise.get_reply()
                answer = reply.get_value('answer')
                # Set local description
                promise = Gst.Promise.new()
                self.webrtcbin.emit('set-local-description', answer, promise)
                promise.interrupt()

                answer_text = answer.sdp.as_text()
                await self.start(answer_text)
            else:
                raise Exception("Invalid JSEP")

        elif 'ice' in jsep:
            ice = jsep['ice']
            candidate = ice['candidate']
            sdpMLineIndex = ice['sdpMLineIndex']
            self.webrtcbin.emit('add-ice-candidate', sdpMLineIndex, candidate)



import logging
from abc import ABC, abstractmethod

from aiortc import (
    RTCPeerConnection,
    RTCSessionDescription,
)

from .session import JanusSession
from .message_transaction import MessageTransaction


logger = logging.getLogger(__name__)


class JanusPlugin(ABC):
    """Base class to inherit when implementing a plugin"""

    name: str = "janus.plugin.base.dummy"
    """Plugin name

    Must override to match plugin name in Janus server.
    """

    __id: str
    """Plugin handle ID. Given by Janus."""

    __session: JanusSession
    """Session instance this plugin is created from."""

    _pc: RTCPeerConnection
    """A WebRTC PeerConnection. A plugin handle is expected to have
    only 1 PC.
    """

    def __init__(self) -> None:
        self.__id = None
        self._pc = RTCPeerConnection()

    @property
    def id(self) -> int:
        return self.__id

    async def attach(self, session: JanusSession):
        if self.__id:
            raise Exception(f"Plugin already attached to session ({self.__session})")

        self.__session = session
        self.__id = await session.attach_plugin(self)

    async def destroy(self):
        """Destroy plugin handle"""

        message_transaction = await self.send({"janus": "detach"})
        await message_transaction.get()
        await message_transaction.done()
        self.__session.detach_plugin(self)

    def __sanitize_message(self, message: dict) -> None:
        if "handle_id" in message:
            logger.warn(
                f"Should not set handle_id ({message['handle_id']}). Overriding."
            )
            del message["handle_id"]

    async def send(
        self,
        message: dict,
    ) -> MessageTransaction:
        """Send raw message to plugin

        Will auto attach plugin ID to the message.

        :param message: JSON serializable dictionary to send
        :return: Synchronous reply from server
        """

        self.__sanitize_message(message=message)

        return await self.__session.send(message, handle_id=self.__id)

    @abstractmethod
    async def on_receive(self, response: dict):
        """Handle asynchronous events from Janus"""
        pass

    async def create_jsep(self, pc: RTCPeerConnection, trickle: bool = False) -> dict:
        return {
            "sdp": pc.localDescription.sdp,
            "trickle": trickle,
            "type": pc.localDescription.type,
        }

    async def on_receive_jsep(self, jsep: dict):
        if self._pc and self._pc.signalingState != "closed":
            await self._pc.setRemoteDescription(
                RTCSessionDescription(sdp=jsep["sdp"], type=jsep["type"])
            )

    async def trickle(self, sdpMLineIndex, candidate):
        """Send WebRTC candidates to Janus

        :param sdpMLineIndex: (I don't know what is this)
        :param candidate: Candidate payload. (I got it from WebRTC instance callback)
        """

        candidate_payload = dict()
        if candidate:
            candidate_payload = {
                "sdpMLineIndex": sdpMLineIndex,
                "candidate": candidate,
            }
        else:
            # Reference: https://janus.conf.meetecho.com/docs/rest.html
            # - a null candidate or a completed JSON object to notify the end of the candidates.
            # TODO: test it
            candidate_payload = None

        # await self.send({"janus": "trickle", "candidate": candidate_payload})
        await self.send({"janus": "trickle", "candidate": candidate_payload})
        # TODO: Implement sending an array of candidates



import logging
from typing import Any
import asyncio
import json
import traceback

import websockets

from .transport import JanusTransport


logger = logging.getLogger(__name__)


class JanusTransportWebsocket(JanusTransport):
    """Janus transport through HTTP

    Manage Sessions and Transactions
    """

    ws: websockets.WebSocketClientProtocol
    subprotocol: str

    def __init__(self, **kwargs: dict):
        super().__init__(**kwargs)

        self._connected = False
        if "subprotocol" in kwargs:
            self.subprotocol = kwargs["subprotocol"]
        else:
            self.subprotocol = "janus-protocol"

    async def _connect(self, **kwargs: Any) -> None:
        """Connect to server

        All extra keyword arguments will be passed to websockets.connect
        """

        logger.info(f"Connecting to: {self.base_url}")

        self.ws = await websockets.connect(
            self.base_url,
            subprotocols=[websockets.Subprotocol(self.subprotocol)],
            **kwargs,
        )
        self.receive_message_task = asyncio.create_task(self.receive_message())
        self.receive_message_task.add_done_callback(self.receive_message_done_cb)

        self.connected = True
        logger.info("Connected")

    async def _disconnect(self) -> None:
        logger.info("Disconnecting")
        self.receive_message_task.cancel()
        # This wait might not be useful, but leaving it here
        await asyncio.wait([self.receive_message_task])
        await self.ws.close()
        self.connected = False
        logger.info("Disconnected")

    def receive_message_done_cb(self, task: asyncio.Task, context=None) -> None:
        try:
            # Check if any exceptions are raised
            # If it's CancelledError or InvalidStateError exception then they will be raised
            # else the exception in task will be returned
            exception = task.exception()
            if exception:
                logger.error(
                    "".join(
                        traceback.format_exception(
                            etype=type(exception),
                            value=exception,
                            tb=exception.__traceback__,
                        )
                    )
                )
        except asyncio.CancelledError:
            logger.info("Receive message task ended")
        except asyncio.InvalidStateError:
            logger.info("receive_message_done_cb called with invalid state")

        self.connected = False

    async def receive_message(self) -> None:
        if not self.ws:
            raise Exception("Not connected to server.")

        async for message_raw in self.ws:
            response = json.loads(message_raw)

            await self.receive(response)

    async def _send(
        self,
        message: dict,
    ) -> None:
        if not self.connected:
            raise Exception("Must connect before any communication.")

        await self.ws.send(json.dumps(message))


def protocol_matcher(base_url: str):
    return base_url.startswith(("ws://", "wss://"))


JanusTransport.register_transport(
    protocol_matcher=protocol_matcher, transport_cls=JanusTransportWebsocket
)



import uuid
from typing import Dict, Union, List
import logging

from .transport import JanusTransport
from .transport_http import JanusTransportHTTP
from .message_transaction import is_subset


logger = logging.getLogger(__name__)


"""
# Take note to enable admin API with websockets in Janus, for example:
# admin: {
#         admin_ws = true                         # Whether to enable the Admin API WebSockets API
#         admin_ws_port = 7188                    # Admin API WebSockets server port, if enabled
#         #admin_ws_interface = "eth0"            # Whether we should bind this server to a specific interface only
#         #admin_ws_ip = "192.168.0.1"            # Whether we should bind this server to a specific IP address only
#         admin_wss = true                        # Whether to enable the Admin API secure WebSockets
#         admin_wss_port = 7989                   # Admin API WebSockets server secure port, if enabled
#         #admin_wss_interface = "eth0"           # Whether we should bind this server to a specific interface only
#         #admin_wss_ip = "192.168.0.1"           # Whether we should bind this server to a specific IP address only
#         #admin_ws_acl = "127.,192.168.0."       # Only allow requests coming from this comma separated list of addresses
# }

When it comes to authentication:
- If stored-token auth (token) is enabled, all requests will require auth.
  At this point, the token can only be created through admin add token API.
- Admin API uses it's own auth method (admin_secret)
- If shared static token auth (api_secret) is enabled, all requests can choose to use this
  auth instead of token auth.
"""


class JanusAdminMonitorClient:
    """
    An Admin/Monitor API that can be used to ask Janus for more specific information related to sessions and handles.
    """

    __transport: JanusTransport
    __admin_secret: str

    def __init__(
        self,
        base_url: str,
        admin_secret: str,
        api_secret: str = None,
        token: str = None,
    ):
        self.__transport = JanusTransport.create_transport(
            base_url=base_url,
            api_secret=api_secret,
            token=token,
            config={"subprotocol": "janus-admin-protocol"},
        )
        self.__admin_secret = admin_secret

    def __str__(self):
        return f"Admin/Monitor ({self.__transport.base_url}) {self}"

    async def connect(self) -> None:
        """Initialize resources"""

        await self.__transport.connect()

    async def disconnect(self) -> None:
        """Release resources"""

        await self.__transport.disconnect()

    async def send_wrapper(
        self,
        message: dict,
        matcher: dict = {},
        jsep: dict = {},
        timeout: Union[float, None] = 15,
        authorize: bool = True,
    ) -> dict:
        def function_matcher(message: dict):
            return is_subset(message, matcher) or is_subset(
                message,
                {
                    "janus": "error",
                    "error": {
                        "code": None,
                        "reason": None,
                    },
                },
            )

        full_message = message
        if jsep:
            full_message = {**message, "jsep": jsep}

        if authorize:
            full_message["admin_secret"] = self.__admin_secret

        message_transaction = await self.__transport.send(
            message=full_message,
        )
        response = await message_transaction.get(
            matcher=function_matcher,
            timeout=timeout,
        )
        await message_transaction.done()

        return response

    async def ping(self) -> Dict:
        """A simple ping/pong mechanism with server. Doesn't require admin secret."""

        return await self.send_wrapper(
            message={"janus": "ping"},
            matcher={"janus": "pong"},
            authorize=False,
        )

    async def info(self) -> Dict:
        """
        Get server info. Gets the same info as transport info API.
        Doesn't require admin secret.
        """

        if isinstance(self.__transport, JanusTransportHTTP):
            return await self.__transport.info()
        else:
            return await self.send_wrapper(
                message={"janus": "info"},
                matcher={"janus": "server_info"},
                authorize=False,
            )

    async def loops_info(self) -> List:
        """
        Returns a summary of how many handles each static event loop is
        currently responsible for, in case static event loops are
        in use (returns an empty array otherwise).
        """

        response = await self.send_wrapper(
            message={"janus": "loops_info"}, matcher={"janus": "success"}
        )
        return response["loops"]

    # Configuration related requests

    async def get_settings(self) -> Dict:
        """
        Gets the current value for the settings that can be modified at
        runtime via the Admin API.
        """

        response = await self.send_wrapper(
            message={"janus": "get_status"},
            matcher={"janus": "success", "status": {}},
        )
        return response["status"]

    async def set_session_timeout(self, session_timeout: int) -> int:
        """
        Change global session timeout value in Janus.
        Returns the value that it is set to.
        """

        response = await self.send_wrapper(
            message={"janus": "set_session_timeout", "timeout": session_timeout},
            matcher={"janus": "success", "timeout": None},
        )
        return response["timeout"]

    async def set_log_level(self, log_level: int) -> int:
        """
        Change the log level in Janus.
        Returns the value that it is set to.
        """

        response = await self.send_wrapper(
            message={"janus": "set_log_level", "level": log_level},
            matcher={"janus": "success", "level": None},
        )
        return response["level"]

    async def set_log_timestamps(self, log_timestamps: bool) -> bool:
        """
        Selectively enable/disable adding a timestamp to all log lines
        Janus writes on the console and/or to file.
        Returns the value that it is set to.
        """

        response = await self.send_wrapper(
            message={"janus": "set_log_timestamps", "timestamps": log_timestamps},
            matcher={"janus": "success", "log_timestamps": None},
        )
        return response["log_timestamps"]

    async def set_log_colors(self, log_colors: bool) -> bool:
        """
        Selectively enable/disable using colors in all log lines
        Janus writes on the console and/or to file.
        Returns the value that it is set to.
        """

        response = await self.send_wrapper(
            message={"janus": "set_log_colors", "colors": log_colors},
            matcher={"janus": "success", "log_colors": None},
        )
        return response["log_colors"]

    async def set_locking_debug(self, locking_debug: bool) -> bool:
        """
        Selectively enable/disable a live debugging of the locks in
        Janus on the fly (useful if you're experiencing deadlocks
        and want to investigate them).
        Returns the value that it is set to.
        """

        response = await self.send_wrapper(
            message={"janus": "set_locking_debug", "debug": locking_debug},
            matcher={"janus": "success", "locking_debug": None},
        )
        return response["locking_debug"]

    async def set_refcount_debug(self, refcount_debug: bool) -> bool:
        """
        Selectively enable/disable a live debugging of the reference
        counters in Janus on the fly (useful if you're experiencing
        memory leaks in the Janus structures and want to investigate them).
        Returns the value that it is set to.
        """

        response = await self.send_wrapper(
            message={"janus": "set_refcount_debug", "debug": refcount_debug},
            matcher={"janus": "success", "refcount_debug": None},
        )
        return response["refcount_debug"]

    async def set_libnice_debug(self, libnice_debug: bool) -> bool:
        """
        Selectively enable/disable libnice debugging.
        Returns the value that it is set to.
        """

        response = await self.send_wrapper(
            message={"janus": "set_libnice_debug", "debug": libnice_debug},
            matcher={"janus": "success", "libnice_debug": None},
        )
        return response["libnice_debug"]

    async def set_min_nack_queue(self, min_nack_queue: int) -> int:
        """
        Change the value of the min NACK queue window.
        Returns the value that it is set to.
        """

        response = await self.send_wrapper(
            message={"janus": "set_min_nack_queue", "min_nack_queue": min_nack_queue},
            matcher={"janus": "success", "min_nack_queue": None},
        )
        return response["min_nack_queue"]

    async def set_no_media_timer(self, no_media_timer: int) -> int:
        """
        Change the value of the no-media timer property.
        Returns the value that it is set to.
        """

        response = await self.send_wrapper(
            message={"janus": "set_no_media_timer", "no_media_timer": no_media_timer},
            matcher={"janus": "success", "no_media_timer": None},
        )
        return response["no_media_timer"]

    async def set_slowlink_threshold(self, slowlink_threshold: int) -> int:
        """
        Change the value of the slowlink-threshold property.
        Returns the value that it is set to.
        """

        response = await self.send_wrapper(
            message={
                "janus": "set_slowlink_threshold",
                "slowlink_threshold": slowlink_threshold,
            },
            matcher={"janus": "success", "slowlink_threshold": None},
        )
        return response["slowlink_threshold"]

    # Token related requests

    async def list_tokens(self) -> List:
        """
        List the existing tokens
        (only available if you enabled the Stored token based authentication mechanism);
        """

        response = await self.send_wrapper(
            message={"janus": "list_tokens"},
            matcher={"janus": "success", "data": {"tokens": None}},
        )
        return response["data"]["tokens"]

    async def add_token(self, token: str = uuid.uuid4().hex, plugins: list = []) -> str:
        """
        Add a valid token
        (only available if you enabled the Stored token based authentication mechanism)

        Ok to add the same token repeatedly.
        Plugin permissions provided in input will be added to existing permissions.
        Providing empty plugin permissions will allow access to all plugins.
        """

        success_matcher = {"janus": "success", "data": {"plugins": None}}
        response = await self.send_wrapper(
            message={
                "janus": "add_token",
                "token": token,
                "plugins": plugins,
            },
            matcher=success_matcher,
        )

        if not is_subset(response, success_matcher):
            raise Exception("Fail to add token")

        return token

    async def remove_token(self, token: str) -> bool:
        """
        Remove a token
        (only available if you enabled the Stored token based authentication mechanism)

        Will fail if the token is not already added.
        """

        success_matcher = {"janus": "success"}
        response = await self.send_wrapper(
            message={
                "janus": "remove_token",
                "token": token,
            },
            matcher=success_matcher,
        )

        if not is_subset(response, success_matcher):
            raise Exception("Fail to remove token")

        return True

    async def allow_token(self, token: str, plugins: list) -> List:
        """
        Give a token access to a plugin
        (only available if you enabled the Stored token based authentication mechanism)

        Ok to allow the same token permissions repeatedly.
        Plugin permissions provided in input will be added to existing permissions.
        Empty permission list is not accepted.
        Invalid plugin permissions are not accepted.
        """

        if not plugins:
            raise Exception("plugins should be non-empty array")

        success_matcher = {"janus": "success", "data": {"plugins": None}}
        response = await self.send_wrapper(
            message={
                "janus": "allow_token",
                "token": token,
                "plugins": plugins,
            },
            matcher=success_matcher,
        )

        if not is_subset(response, success_matcher):
            raise Exception("Fail to allow token")

        return response["data"]["plugins"]

    async def disallow_token(self, token: str, plugins: list):
        """
        Remove a token access from a plugin
        (only available if you enabled the Stored token based authentication mechanism)

        Ok to disallow the same token permissions repeatedly.
        Plugin permissions provided in input will be removed from existing permissions.
        Empty permission list is not accepted.
        Invalid plugin permissions are not accepted.
        """

        if not plugins:
            raise Exception("plugins should be non-empty array")

        success_matcher = {"janus": "success", "data": {"plugins": None}}
        response = await self.send_wrapper(
            message={
                "janus": "disallow_token",
                "token": token,
                "plugins": plugins,
            },
            matcher=success_matcher,
        )

        if not is_subset(response, success_matcher):
            raise Exception("Fail to disallow token")

        return response["data"]["plugins"]



import asyncio
import logging

from .plugin_base import JanusPlugin
from aiortc import RTCPeerConnection, RTCSessionDescription, VideoStreamTrack
from aiortc.contrib.media import MediaPlayer, MediaRecorder

logger = logging.getLogger(__name__)


class JanusEchoTestPlugin(JanusPlugin):
    """Janus EchoTest plugin implementation"""

    name = "janus.plugin.echotest"
    __pc: RTCPeerConnection
    __recorder: MediaRecorder
    __webrtcup_event: asyncio.Event

    def __init__(self) -> None:
        super().__init__()

        self.__webrtcup_event = asyncio.Event()

    async def on_receive(self, response: dict):
        if "jsep" in response:
            await self.on_receive_jsep(jsep=response["jsep"])

        janus_code = response["janus"]

        if janus_code == "media":
            if response["receiving"]:
                # It's ok to start multiple times, only the track that
                # has not been started will start
                await self.__recorder.start()

        if janus_code == "webrtcup":
            self.__webrtcup_event.set()

        if janus_code == "event":
            plugin_data = response["plugindata"]["data"]

            if plugin_data["echotest"] != "event":
                # This plugin will only get events
                logger.error(f"Invalid response: {response}")
                return

            if "result" in plugin_data:
                if plugin_data["result"] == "ok":
                    # Successful start stream request. Do nothing.
                    pass

                if plugin_data["result"] == "done":
                    # Stream ended. Ok to close PC multiple times.
                    if self.__pc:
                        await self.__pc.close()
                    # Ok to stop recording multiple times.
                    if self.__recorder:
                        await self.__recorder.stop()

            if "errorcode" in plugin_data:
                logger.error(f"Plugin Error: {response}")

    async def wait_webrtcup(self) -> None:
        await self.__webrtcup_event.wait()
        self.__webrtcup_event.clear()

    async def on_receive_jsep(self, jsep: dict):
        if self.__pc and self.__pc.signalingState != "closed":
            await self.__pc.setRemoteDescription(
                RTCSessionDescription(sdp=jsep["sdp"], type=jsep["type"])
            )

    async def start(self, play_from: str, record_to: str = ""):
        self.__pc = RTCPeerConnection()

        player = MediaPlayer(play_from)

        # configure media
        if player and player.audio:
            self.__pc.addTrack(player.audio)

        if player and player.video:
            self.__pc.addTrack(player.video)
        else:
            self.__pc.addTrack(VideoStreamTrack())

        if record_to:
            self.__recorder = MediaRecorder(record_to)

            @self.__pc.on("track")
            async def on_track(track):
                logger.info("Track %s received" % track.kind)
                if track.kind == "video":
                    self.__recorder.addTrack(track)
                if track.kind == "audio":
                    self.__recorder.addTrack(track)

        # send offer
        await self.__pc.setLocalDescription(await self.__pc.createOffer())

        message = {"janus": "message"}
        body = {
            "audio": bool(player.audio),
            # "audiocodec" : "<optional codec name; only used when creating a PeerConnection>",
            "video": bool(player.video),
            # "videocodec" : "<optional codec name; only used when creating a PeerConnection>",
            # "videoprofile" : "<optional codec profile to force; only used when creating a PeerConnection, only valid for VP9 (0 or 2) and H.264 (e.g., 42e01f)>",
            # "bitrate" : <numeric bitrate value>,
            # "record" : true|false,
            # "filename" : <base path/filename to use for the recording>,
            # "substream" : <substream to receive (0-2), in case simulcasting is enabled>,
            # "temporal" : <temporal layers to receive (0-2), in case simulcasting is enabled>,
            # "svc" : true|false,
            # "spatial_layer" : <spatial layer to receive (0-2), in case SVC is enabled>,
            # "temporal_layer" : <temporal layers to receive (0-2), in case SVC is enabled>
        }
        message["body"] = body
        message["jsep"] = {
            "sdp": self.__pc.localDescription.sdp,
            "trickle": False,
            "type": self.__pc.localDescription.type,
        }

        message_transaction = await self.send(message)
        response = await message_transaction.get()
        await message_transaction.done()

        # Immediately apply answer if it's found
        if "jsep" in response:
            await self.on_receive_jsep(jsep=response["jsep"])

    async def close_stream(self):
        """Close stream

        This should cause the stream to stop and a done event to be received.
        """
        if self.__pc:
            await self.__pc.close()

        if self.__recorder:
            await self.__recorder.stop()



from abc import ABC, abstractmethod
import asyncio
from typing import TYPE_CHECKING, List, Dict
import logging

# import uuid
import json

from .message_transaction import MessageTransaction

if TYPE_CHECKING:
    from .session import JanusSession

logger = logging.getLogger(__name__)


class JanusTransport(ABC):
    """Janus transport protocol interface

    Manage Sessions and Transactions
    """

    __transport_implementation: List[tuple] = []

    __base_url: str
    __api_secret: str
    __token: str
    __message_transaction: Dict[str, MessageTransaction]
    __sessions: Dict[int, "JanusSession"]
    __connect_lock: asyncio.Lock
    connected: bool
    """Must set this property when connected or disconnected"""

    @abstractmethod
    async def _send(self, message: Dict) -> None:
        """Really sends the message. Doesn't return a response"""
        pass

    @abstractmethod
    async def _connect(self) -> None:
        pass

    @abstractmethod
    async def _disconnect(self) -> None:
        pass

    async def info(self) -> Dict:
        """Get info of Janus server. Will be overridden for HTTP."""
        message_transaction = await self.send({"janus": "info"})
        response = await message_transaction.get()
        await message_transaction.done()
        return response

    async def ping(self) -> Dict:
        message_transaction = await self.send(
            {"janus": "ping"},
            # response_handler=lambda res: res if res["janus"] == "pong" else None,
        )
        response = await message_transaction.get(matcher={"janus": "pong"}, timeout=15)
        await message_transaction.done()
        return response

    async def dispatch_session_created(self, session_id: int) -> None:
        """Override this method to get session created event"""
        pass

    async def dispatch_session_destroyed(self, session_id: int) -> None:
        """Override this method to get session destroyed event"""
        pass

    def __init__(self, base_url: str, api_secret: str = None, token: str = None, **kwargs: dict):
        """Create connection instance

        :param base_url: Janus server address
        :param api_secret: (optional) API key for shared static secret authentication
        :param token: (optional) Token for shared token based authentication
        """

        self.__base_url = base_url.rstrip("/")
        self.__api_secret = api_secret
        self.__token = token
        self.__message_transaction = dict()
        self.__sessions = dict()
        self.__connect_lock = asyncio.Lock()
        self.connected = False

    # def __del__(self):
    #     asyncio.run(asyncio.create_task(self.disconnect()))

    @property
    def base_url(self) -> str:
        return self.__base_url

    # async def put_response(self, transaction_id: int, response: dict) -> None:
    #     logger.info(f"Received: {response}")
    #     await self.__transactions[transaction_id].put(response)

    async def connect(self) -> None:
        """Initialize resources"""
        async with self.__connect_lock:
            if not self.connected:
                await self._connect()

                self.connected = True

    async def disconnect(self) -> None:
        """Release resources"""
        async with self.__connect_lock:
            if self.connected:
                await self._disconnect()

                self.connected = False

    def __sanitize_message(self, message: Dict) -> None:
        if "janus" not in message:
            raise Exception('Must set "janus" field')

        if "transaction" in message:
            logger.warn(
                f"Should not set transaction ({message['transaction']}). Overriding."
            )
            del message["transaction"]

    async def send(
        self,
        message: Dict,
        session_id: int = None,
        handle_id: int = None,
    ) -> MessageTransaction:
        """Send message to server

        :param message: JSON serializable dictionary to send

        :returns: Synchronous response from Janus server

        """

        self.__sanitize_message(message=message)

        # Create transaction
        message_transaction = MessageTransaction()
        self.__message_transaction[message_transaction.id] = message_transaction
        message["transaction"] = message_transaction.id

        # Delete itself if done is called
        async def message_transaction_on_done():
            del self.__message_transaction[message_transaction.id]

        message_transaction.on_done = message_transaction_on_done

        # Authentication
        if self.__api_secret is not None:
            message["apisecret"] = self.__api_secret
        if self.__token is not None:
            message["token"] = self.__token

        # IDs
        if session_id is not None:
            message["session_id"] = session_id
        if handle_id is not None:
            message["handle_id"] = handle_id

        # Send the message
        message_json = json.dumps(message)
        logger.info(f"Send: {message_json}")
        await self._send(message=message)

        return message_transaction

    async def receive(self, response: dict) -> None:
        logger.info(f"Received: {response}")
        # First try transaction handlers
        if "transaction" in response:
            transaction_id = response["transaction"]

            if transaction_id in self.__message_transaction:
                self.__message_transaction[transaction_id].put_msg(message=response)
                return

        # If the response was not "eaten" by the transaction, then dispatch it
        if "session_id" in response:
            session_id = response["session_id"]
            # This is response for session or plugin handle
            if session_id in self.__sessions:
                await self.__sessions[session_id].on_receive(response)
            else:
                logger.warning(
                    f"Got response for session but session not found."
                    f"Session ID: {session_id} Unhandeled response: {response}"
                )
        else:
            # No handler found for response
            logger.info(f"Response dropped: {response}")

    async def create_session(self, session: "JanusSession") -> int:
        """Create Janus Session"""

        message_transaction = await self.send({"janus": "create"})
        response = await message_transaction.get()
        await message_transaction.done()

        # Extract session ID
        session_id = int(response["data"]["id"])

        # Register session
        self.__sessions[session_id] = session

        await self.dispatch_session_created(session_id=session_id)

        return session_id

    # Don't call this from client object, call destroy from session instead
    async def destroy_session(self, session_id: int) -> None:
        del self.__sessions[session_id]

        await self.dispatch_session_destroyed(session_id=session_id)

        # Also release transport resources if this is the last session
        if len(self.__sessions) == 0:
            await self.disconnect()

    @staticmethod
    def register_transport(protocol_matcher, transport_cls: "JanusTransport") -> None:
        JanusTransport.__transport_implementation.append(
            (protocol_matcher, transport_cls)
        )

    @staticmethod
    def create_transport(
        base_url: str, api_secret: str = None, token: str = None, config: Dict = {}
    ) -> "JanusTransport":
        # Get matching results
        matching_results = []
        for transport_implementation in JanusTransport.__transport_implementation:
            protocol_matcher = transport_implementation[0]
            matching_results.append(protocol_matcher(base_url))

        total_matched = sum(map(bool, matching_results))

        # Cannot have more than 1 match
        if total_matched > 1:
            logger.info(JanusTransport.__transport_implementation)
            logger.info(matching_results)
            raise Exception("Matched to more than 1 protocol")
        elif total_matched == 0:
            logger.info(JanusTransport.__transport_implementation)
            logger.info(matching_results)
            raise Exception("No protocol matched")

        for index, result in enumerate(matching_results):
            if result:
                transport_protocol = JanusTransport.__transport_implementation[index][1]
                return transport_protocol(
                    base_url=base_url, api_secret=api_secret, token=token, **config
                )



import logging
from enum import Enum
from typing import List

from aiortc import (
    RTCPeerConnection,
    RTCSessionDescription,
    VideoStreamTrack,
    MediaStreamTrack,
)
from aiortc.contrib.media import MediaPlayer, MediaRecorder

from .plugin_base import JanusPlugin
from .message_transaction import is_subset

logger = logging.getLogger(__name__)


class AllowedAction(Enum):
    ENABLE = "enable"
    DISABLE = "disable"
    ADD = "add"
    REMOVE = "remove"


class JanusVideoRoomPlugin(JanusPlugin):
    """Janus VideoRoom plugin implementation

    Implements API to interact with VideoRoom plugin.

    Each plugin object is expected to have only 1 PeerConnection
    """

    name = "janus.plugin.videoroom"  #: Plugin name
    __player: MediaPlayer
    __recorder: MediaRecorder

    # def __init__(self, *args, **kwargs):
    #     super().__init__(*args, **kwargs)

    async def on_receive(self, response: dict):
        """Handle asynchronous messages"""

        janus_code = response["janus"]

        if janus_code == "media":
            if response["receiving"]:
                # It's ok to start multiple times, only the track that
                # has not been started will start
                if self.__recorder:
                    await self.__recorder.start()

        if janus_code == "event":
            logger.info(f"Event response: {response}")
            # if "plugindata" in response:
            #     if response["plugindata"]["data"]["videoroom"] == "attached":
            #         # Subscriber attached
            #         self.joined_event.set()
            #     elif response["plugindata"]["data"]["videoroom"] == "joined":
            #         # Participant joined (joined as publisher but may not publish)
            #         self.joined_event.set()
        else:
            logger.info(f"Unimplemented response handle: {response}")

        # VideoRoom plugin doesn't send JSEP asynchronously

    async def send_wrapper(self, message: dict, matcher: dict, jsep: dict = {}) -> dict:
        def function_matcher(message: dict):
            return (
                is_subset(message, matcher)
                or is_subset(
                    message,
                    {
                        "janus": "success",
                        "plugindata": {
                            "plugin": self.name,
                            "data": {
                                "videoroom": "event",
                                "error_code": None,
                                "error": None,
                            },
                        },
                    },
                )
                or is_subset(
                    message,
                    {
                        "janus": "event",
                        "plugindata": {
                            "plugin": self.name,
                            "data": {
                                "videoroom": "event",
                                "error_code": None,
                                "error": None,
                            },
                        },
                    },
                )
                or is_subset(message, {"janus": "error", "error": {}})
            )

        full_message = message
        if jsep:
            full_message = {**message, "jsep": jsep}

        message_transaction = await self.send(
            message=full_message,
        )
        response = await message_transaction.get(matcher=function_matcher, timeout=15)
        await message_transaction.done()

        if is_subset(response, {"janus": "error", "error": {}}):
            raise Exception(f"Janus error: {response}")

        return response

    async def create_room(self, room_id: int, configuration: dict = {}) -> bool:
        """Create a room.

        Refer to documentation for description of parameters.
        https://janus.conf.meetecho.com/docs/videoroom.html
        """

        success_matcher = {
            "janus": "success",
            "plugindata": {"plugin": self.name, "data": {"videoroom": "created"}},
        }
        response = await self.send_wrapper(
            message={
                "janus": "message",
                "body": {
                    "request": "create",
                    "room": room_id,
                    **configuration,
                },
            },
            matcher=success_matcher,
        )

        return is_subset(response, success_matcher)

    async def destroy_room(
        self, room_id: int, secret: str = "", permanent: bool = False
    ) -> bool:
        """Destroy a room.

        All other participants in the room will also get the "destroyed" event.
        """

        success_matcher = {
            "janus": "success",
            "plugindata": {
                "plugin": self.name,
                "data": {"videoroom": "destroyed", "room": room_id},
            },
        }
        response = await self.send_wrapper(
            message={
                "janus": "message",
                "body": {
                    "request": "destroy",
                    "room": room_id,
                    "secret": secret,
                    "permanent": permanent,
                },
            },
            matcher=success_matcher,
        )

        return is_subset(response, success_matcher)

    async def edit(
        self,
        room_id: int,
        secret: str = "",
        new_description: str = "",
        new_secret: str = "",
        new_pin: str = "",
        new_is_private: bool = False,
        new_require_pvtid: bool = False,
        new_bitrate: int = None,
        new_fir_freq: int = None,
        new_publishers: int = 3,
        new_lock_record: bool = False,
        new_rec_dir: str = None,
        permanent: bool = False,
    ) -> bool:
        """Edit a room."""

        success_matcher = {
            "janus": "success",
            "plugindata": {
                "plugin": self.name,
                "data": {"videoroom": "edited", "room": room_id},
            },
        }

        body = {
            "request": "edit",
            "room": room_id,
            "secret": secret,
            "new_description": new_description,
            "new_secret": new_secret,
            "new_pin": new_pin,
            "new_is_private": new_is_private,
            "new_require_pvtid": new_require_pvtid,
            "new_publishers": new_publishers,
            "new_lock_record": new_lock_record,
            "permanent": permanent,
        }
        if new_bitrate:
            body["new_bitrate"] = new_bitrate
        if new_fir_freq:
            body["new_fir_freq"] = new_fir_freq
        if new_rec_dir:
            body["new_rec_dir"] = new_rec_dir

        response = await self.send_wrapper(
            message={
                "janus": "message",
                "body": body,
            },
            matcher=success_matcher,
        )

        return is_subset(response, success_matcher)

    async def exists(self, room_id: int) -> bool:
        """Check if a room exists."""

        success_matcher = {
            "janus": "success",
            "plugindata": {
                "plugin": self.name,
                "data": {"videoroom": "success", "room": room_id, "exists": None},
            },
        }
        response = await self.send_wrapper(
            message={
                "janus": "message",
                "body": {
                    "request": "exists",
                    "room": room_id,
                },
            },
            matcher=success_matcher,
        )

        return (
            is_subset(response, success_matcher)
            and response["plugindata"]["data"]["exists"]
        )

    async def allowed(
        self,
        room_id: int,
        secret: str = "",
        action: AllowedAction = AllowedAction.ENABLE,
        allowed: List[str] = [],
    ) -> bool:
        """Configure ACL of a room."""

        success_matcher = {
            "janus": "success",
            "plugindata": {
                "plugin": self.name,
                "data": {"videoroom": "success", "room": room_id, "allowed": None},
            },
        }
        response = await self.send_wrapper(
            message={
                "janus": "message",
                "body": {
                    "request": "allowed",
                    "room": room_id,
                    "secret": secret,
                    "action": action.value,
                    "allowed": allowed,
                },
            },
            matcher=success_matcher,
        )

        return is_subset(response, success_matcher)

    async def kick(
        self,
        room_id: int,
        id: str,
        secret: str = "",
    ) -> bool:
        """Kick a participant by ID.

        Only works for room administrators (i.e. you created the room).
        """

        success_matcher = {
            "janus": "success",
            "plugindata": {
                "plugin": self.name,
                "data": {"videoroom": "success"},
            },
        }
        response = await self.send_wrapper(
            message={
                "janus": "message",
                "body": {
                    "request": "kick",
                    "room": room_id,
                    "secret": secret,
                    "id": id,
                },
            },
            matcher=success_matcher,
        )

        return is_subset(response, success_matcher)

    async def moderate(
        self,
        room_id: int,
        id: str,
        mid: str,
        mute: bool,
        secret: str = "",
    ) -> bool:
        """Moderate a participant by ID.

        Only works for room administrators (i.e. you created the room).
        """

        success_matcher = {
            "janus": "success",
            "plugindata": {
                "plugin": self.name,
                "data": {"videoroom": "success"},
            },
        }
        response = await self.send_wrapper(
            message={
                "janus": "message",
                "body": {
                    "request": "moderate",
                    "room": room_id,
                    "secret": secret,
                    "id": id,
                    "mid": mid,
                    "mute": mute,
                },
            },
            matcher=success_matcher,
        )

        return is_subset(response, success_matcher)

    async def list_room(self) -> List[dict]:
        """List all rooms created.

        If admin_key is included, then private rooms will be listed as well.
        TODO: Find out how to include admin_key.
        """

        success_matcher = {
            "janus": "success",
            "plugindata": {
                "plugin": self.name,
                "data": {"videoroom": "success", "list": None},
            },
        }
        response = await self.send_wrapper(
            message={
                "janus": "message",
                "body": {
                    "request": "list",
                },
            },
            matcher=success_matcher,
        )

        if is_subset(response, success_matcher):
            return response["plugindata"]["data"]["list"]
        else:
            raise Exception(f"Fail to list rooms: {response}")

    async def list_participants(self, room_id: int) -> list:
        """Get participant list in a room

        Get a list of publishers in the room, that are currently publishing.

        :param room_id: List participants in this room
        :return: A list containing the participants. Can be empty.
        """

        success_matcher = {
            "janus": "success",
            "plugindata": {
                "plugin": self.name,
                "data": {
                    "videoroom": "participants",
                    "room": room_id,
                    "participants": None,
                },
            },
        }
        response = await self.send_wrapper(
            message={
                "janus": "message",
                "body": {
                    "request": "listparticipants",
                    "room": room_id,
                },
            },
            matcher=success_matcher,
        )

        if is_subset(response, success_matcher):
            return response["plugindata"]["data"]["participants"]
        else:
            raise Exception(f"Fail to list participants: {response}")

    async def join(
        self,
        room_id: int,
        publisher_id: int = None,
        display_name: str = "",
        token: str = None,
    ) -> bool:
        """Join a room

        A handle can join a room and then do nothing, but this should be called before publishing.
        There is an API to configure and publish at the same time, but it's not implemented yet.

        :param room_id: unique ID of the room to join.
        :param publisher_id: unique ID to register for the publisher; optional, will be chosen by the plugin if missing.
        :param display_name: display name for the publisher; optional.
        :param token: invitation token, in case the room has an ACL; optional.

        :return: True if room is created.
        """

        body = {
            "request": "join",
            "ptype": "publisher",
            "room": room_id,
            "display": display_name,
        }
        if publisher_id:
            body["publisher_id"] = publisher_id
        if token:
            body["token"] = token
        success_matcher = {
            "janus": "event",
            "plugindata": {
                "plugin": self.name,
                "data": {"videoroom": "joined", "room": room_id},
            },
        }

        response = await self.send_wrapper(
            message={
                "janus": "message",
                "body": body,
            },
            matcher=success_matcher,
        )

        return is_subset(response, success_matcher)

    async def leave(self) -> bool:
        """Leave the room. Will unpublish if publishing.

        :return: True if successfully leave.
        """

        success_matcher = {
            "janus": "event",
            "plugindata": {
                "plugin": self.name,
                "data": {"videoroom": "event", "leaving": "ok"},
            },
        }
        response = await self.send_wrapper(
            message={
                "janus": "message",
                "body": {
                    "request": "leave",
                },
            },
            matcher=success_matcher,
        )

        await self._pc.close()

        return is_subset(response, success_matcher)

    async def create_pc(
        self,
        player: MediaPlayer = None,
        recorder: MediaRecorder = None,
        jsep: dict = {},
    ) -> RTCPeerConnection:
        pc = RTCPeerConnection()

        # configure media
        if player:
            if player.audio:
                pc.addTrack(player.audio)

            if player.video:
                pc.addTrack(player.video)
            else:
                # Add dummy video track because aiortc cannot open
                # PC without at least 1 media stream
                pc.addTrack(VideoStreamTrack())
        else:
            # Add dummy video track because aiortc cannot open
            # PC without at least 1 media stream
            pc.addTrack(VideoStreamTrack())

        # Must configure on track event before setRemoteDescription
        if recorder:

            @pc.on("track")
            async def on_track(track: MediaStreamTrack):
                logger.info("Track %s received" % track.kind)
                if track.kind == "video":
                    recorder.addTrack(track)
                if track.kind == "audio":
                    recorder.addTrack(track)

                await recorder.start()

        if jsep:
            await pc.setRemoteDescription(
                RTCSessionDescription(sdp=jsep["sdp"], type=jsep["type"])
            )

        return pc

    async def publish(
        self,
        player: MediaPlayer,
        recorder: MediaRecorder = None,  # Probably no need for recorder in publish
        configuration: dict = {},
    ) -> None:
        """Publish video stream to the room

        Should already have joined a room before this.
        """

        self._pc = await self.create_pc(player=player, recorder=recorder)
        self.__player = player
        self.__recorder = recorder

        # send offer
        await self._pc.setLocalDescription(await self._pc.createOffer())

        body = {
            "request": "publish",
            # "audiocodec" : "<audio codec to prefer among the negotiated ones; optional>",
            # "videocodec" : "<video codec to prefer among the negotiated ones; optional>",
            # "bitrate" : <bitrate cap to return via REMB; optional, overrides the global room value if present>,
            # "record" : <true|false, whether this publisher should be recorded or not; optional>,
            # "filename" : "<if recording, the base path/file to use for the recording files; optional>",
            # "display" : "<display name to use in the room; optional>",
            # "audio_level_average" : "<if provided, overrided the room audio_level_average for this user; optional>",
            # "audio_active_packets" : "<if provided, overrided the room audio_active_packets for this user; optional>",
            # "descriptions" : [      // Optional
            #         {
            #                 "mid" : "<unique mid of a stream being published>",
            #                 "description" : "<text description of the stream (e.g., My front webcam)>"
            #         },
            #         // Other descriptions, if any
            # ]}
            **configuration,
        }

        success_matcher = {
            "janus": "event",
            "plugindata": {
                "plugin": self.name,
                "data": {"videoroom": "event", "configured": "ok"},
            },
        }
        response = await self.send_wrapper(
            message={
                "janus": "message",
                "body": body,
            },
            matcher=success_matcher,
            jsep=await self.create_jsep(pc=self._pc),
        )

        if is_subset(response, success_matcher):
            await self.on_receive_jsep(jsep=response["jsep"])

            return True
        else:
            return False

    async def unpublish(self) -> bool:
        """Stop publishing.

        :return: True if successfully unpublished.
        """

        success_matcher = {
            "janus": "event",
            "plugindata": {
                "plugin": self.name,
                "data": {"videoroom": "event", "unpublished": "ok"},
            },
        }
        response = await self.send_wrapper(
            message={
                "janus": "message",
                "body": {
                    "request": "unpublish",
                },
            },
            matcher=success_matcher,
        )

        await self._pc.close()

        return is_subset(response, success_matcher)

    # TODO: Implement "configure", 'joinandconfigure", "rtp_forward", "stop_rtp_forward", "listforwarders", "enable_recording"

    async def subscribe_and_start(
        self,
        room_id: int,
        recorder: MediaRecorder,
        stream: dict,
        use_msid: bool = False,
        autoupdate: bool = True,
        private_id: int = None,
        # streams: List = [],
    ) -> bool:
        """Subscribe to a feed. Only supporting subscribe to 1 stream.

        :param room_id: Room ID containing the feed. The same ID that
            you would use to join the room.
        :param recorder: AIORTC MediaRecorder
        :param stream: Configuration of the stream to subscribe to. Minimum should have
            a feed ID.
        :param use_msid: whether subscriptions should include an msid that references the publisher; false by default.
        :param autoupdate: whether a new SDP offer is sent automatically when a subscribed publisher leaves; true by default.
        :param private_id: unique ID of the publisher that originated this request; optional, unless mandated by the room configuration.
        """

        body = {
            "request": "join",
            "ptype": "subscriber",
            "room": room_id,
            "use_msid": use_msid,
            "autoupdate": autoupdate,
            "streams": [stream],
        }
        if private_id:
            body["private_id"] = private_id

        success_matcher = {
            "janus": "event",
            "plugindata": {
                "plugin": self.name,
                "data": {"videoroom": "attached", "room": room_id, "streams": []},
            },
        }

        response = await self.send_wrapper(
            message={
                "janus": "message",
                "body": body,
            },
            matcher=success_matcher,
        )

        if not is_subset(response, success_matcher):
            raise Exception("Fail to subscribe.")

        # Successfully attached. Create PeerConnection then start.
        self._pc = await self.create_pc(recorder=recorder, jsep=response["jsep"])
        self.__recorder = recorder
        await self._pc.setLocalDescription(await self._pc.createAnswer())

        return await self.start(
            jsep=await self.create_jsep(self._pc),
        )

    async def unsubscribe(self) -> None:
        """Unsubscribe from the feed"""

        success_matcher = {
            "janus": "event",
            "plugindata": {
                "plugin": self.name,
                "data": {
                    "videoroom": "event",
                    "left": "ok",
                },
            },
        }

        response = await self.send_wrapper(
            message={
                "janus": "message",
                "body": {"request": "leave"},
            },
            matcher=success_matcher,
        )

        await self._pc.close()
        if self.__recorder:
            await self.__recorder.stop()

        return is_subset(response, success_matcher)

    async def start(self, jsep: dict = None) -> bool:
        """Signal WebRTC start."""

        success_matcher = {
            "janus": "event",
            "plugindata": {
                "plugin": self.name,
                "data": {"videoroom": "event", "started": "ok"},
            },
        }

        response = await self.send_wrapper(
            message={
                "janus": "message",
                "body": {"request": "start"},
            },
            matcher=success_matcher,
            jsep=jsep,
        )

        return is_subset(response, success_matcher)

    async def pause(self) -> None:
        """Pause media streaming"""

        message_transaction = await self.send(
            {
                "janus": "message",
                "body": {
                    "request": "pause",
                },
            }
        )
        await message_transaction.get()
        await message_transaction.done()

    # async def handle_jsep(self, jsep):
    #     logger.info(jsep)
    #     if "sdp" in jsep:
    #         sdp = jsep["sdp"]
    #         if jsep["type"] == "answer":
    #             logger.info(f"Received answer:\n{sdp}")

    #             # apply answer
    #             await self.pc.setRemoteDescription(
    #                 RTCSessionDescription(sdp=jsep["sdp"], type=jsep["type"])
    #             )
    #         elif jsep["type"] == "offer":
    #             pass
    #         else:
    #             raise Exception("Invalid JSEP")

    #     elif "ice" in jsep:
    #         ice = jsep["ice"]
    #         candidate = ice["candidate"]
    #         sdpMLineIndex = ice["sdpMLineIndex"]
    #         self.webrtcbin.emit("add-ice-candidate", sdpMLineIndex, candidate)



import asyncio
from typing import Dict, TYPE_CHECKING
import logging

from .transport import JanusTransport
from .message_transaction import MessageTransaction

if TYPE_CHECKING:
    from .plugin_base import JanusPlugin


logger = logging.getLogger(__name__)


class PluginAttachFail(Exception):
    def __init__(self, response: dict) -> None:
        super().__init__(f"Fail to attach plugin: {response['error']}")


class JanusSession:
    """Janus session instance"""

    __id: int
    transport: JanusTransport
    __create_lock: asyncio.Lock
    created: bool

    def __init__(
        self,
        base_url: str = "",
        api_secret: str = None,
        token: str = None,
        transport: JanusTransport = None,
    ):
        self.__id = None
        self.__create_lock = asyncio.Lock()
        self.created = False
        self.plugin_handles: Dict[int, JanusPlugin] = dict()

        if transport:
            self.transport = transport
        else:
            self.transport = JanusTransport.create_transport(
                base_url=base_url,
                api_secret=api_secret,
                token=token,
            )

    def __str__(self):
        return f"Session ({self.__id}) {self}"

    async def _create(self) -> None:
        if not self.transport.connected:
            await self.transport.connect()

        if not self.__id:
            self.__id = await self.transport.create_session(self)

        self.keepalive_task = asyncio.create_task(self.keepalive())
        self.created = True

    async def create(self) -> None:
        """Initialize resources"""
        async with self.__create_lock:
            if not self.created:
                await self._create()

                self.created = True

    async def _destroy(self) -> None:
        message_transaction = await self.send(
            {"janus": "destroy"},
        )
        await message_transaction.get(matcher={"janus": "success"}, timeout=15)
        await message_transaction.done()
        self.keepalive_task.cancel()
        await self.transport.destroy_session(self.__id)
        self.__id = None

    async def destroy(self) -> None:
        """Release resources

        | Should be called when you don't need the session anymore.
        | Plugins from this session should be destroyed before this.
        """
        async with self.__create_lock:
            if self.created:
                await self._destroy()

                self.created = False

    def __sanitize_message(self, message: dict) -> None:
        if "session_id" in message:
            logger.warn(
                f"Should not set session_id ({message['session_id']}). Overriding."
            )
            del message["session_id"]

    async def send(
        self,
        message: dict,
        handle_id: int = None,
    ) -> MessageTransaction:
        self.__sanitize_message(message=message)

        if not self.created:
            await self.create()

        return await self.transport.send(
            message,
            session_id=self.__id,
            handle_id=handle_id,
        )

    # TODO: This is not required if using HTTP REST API, though it
    #       doesn't hurt to still send it.
    async def keepalive(self) -> None:
        # Reference: https://janus.conf.meetecho.com/docs/rest.html
        # A Janus session is kept alive as long as there's no inactivity for 60 seconds
        while True:
            await asyncio.sleep(30)
            message_transaction = await self.send({"janus": "keepalive"})
            await message_transaction.done()

    async def on_receive(self, response: dict):
        if "sender" not in response:
            # This is response for self
            logger.info(f"Async event for session: {response}")
            return

        # This is response for plugin handle
        plugin_id = response["sender"]
        if plugin_id not in self.plugin_handles:
            logger.info(
                f"Got response for plugin handle but handle not found. Handle ID: {plugin_id}"
            )
            logger.info(f"Unhandeled response: {response}")
            return

        await self.plugin_handles[plugin_id].on_receive(response)

    async def attach_plugin(self, plugin: "JanusPlugin") -> int:
        """Create plugin handle for the given plugin type

        :param plugin: Plugin instance with janus_client.JanusPlugin as base class
        """

        def matcher(res):
            return res["janus"] in ["error", "success"]

        message_transaction = await self.send(
            {"janus": "attach", "plugin": plugin.name},
        )
        response = await message_transaction.get(matcher=matcher)
        await message_transaction.done()

        if response["janus"] == "error":
            raise PluginAttachFail(response=response)

        # Extract plugin handle id
        handle_id = int(response["data"]["id"])

        # Register plugin
        self.plugin_handles[handle_id] = plugin

        return handle_id

    def detach_plugin(self, plugin_handle: "JanusPlugin"):
        del self.plugin_handles[plugin_handle.id]
